// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef ORDERINGHEURISTIC_HPP
#define ORDERINGHEURISTIC_HPP

#include "Printable.hpp"
#include "OrderingEntity.hpp"
#include <vector>
#include <sstream>
#include <stdexcept>

class Image;
class GridPoint;
class DlxHead;

static const double NO_FIT  = -2e20;
static const double ONE_FIT = -1e20;
static const double PIECE   =  1e20;

class OrderingHeuristic : public Printable
{
    protected:
        OrderingHeuristic();

    public:
        virtual ~OrderingHeuristic();

        const std::string& getName() const;

        virtual const std::string& getEntityName() const;

        void setEntity(OrderingEntity* entity);

        virtual double gridPointValue(const GridPoint* gp) const;

        /** 
         ** Initialize the selection order enforced by this OrderingHeuristic
         ** (if any).
         **/
        virtual void initSelectionOrder(
                int numFilaGridPoints,
                GridPoint const * const * filaGrid,
                int* selectionOrder) const;

        /** Used by DLX to select the next GridPoint to be filled or piece to
         ** be placed.
         **/
        virtual DlxHead* selectDlx() const;

        /** Used by FILA to select the next GridPoint to be filled.  A set of
         ** Image lists which include all Images that may fit at the selected
         ** GridPoint is also returned.  Although the lists are intended to be
         ** precalculated (this is always the case for the OrderingHeuristics
         ** I've written), the exact list set returned may be specialized to
         ** the situation:
         **
         **    1) The images in each list are always valid images (i.e., they
         **       always fit when the puzzle is empty), so the number of images
         **       in each list will be different depending on the GridPoint
         **       selected.
         **
         **    2) Ordering heuristics that guarantee a predefined hole
         **       selection order will filter out all images in the list of
         **       target hole that conflict with other cells that appear
         **       earlier in this selection order (since they must have
         **       already been filled for the target hole to have been
         **       selected.)
         **
         **    3) If NOF is enabled (see --nof / -n configuration option) then
         **       the returned lists are filtered to exlude any image that
         **       conflicts with neighboring cells that are currently
         **       occupied.
         **
         ** Parameters:
         **    nrs           - The current number of entries in remainingShapes array.
         **
         **    nrp           - The number of remaining pieces.
         **
         **    dlxPristine   - True if and only if the DLX matrix is still an
         **                    accurate representation of the puzzle.
         **                    Explanation:  polycube is setup to initially
         **                    apply DLX until the puzzle is reduced in size
         **                    to some threshold number of remaining pieces,
         **                    (the fila setting) and then place the last fila
         **                    pieces using the FILA algorithm.  If the
         **                    previous piece was placed using DLX, then the
         **                    DLX matrix is still accurate, and so may be
         **                    mined for information as desired.
         **/
        virtual ImageListSet selectFila(int nrp, int nrs, const gridbitfield_t occupancy, bool dlxPristine) const;

        /** Returns true if the selection order is fixed.
         **/
        virtual bool isFixedOrder() const;

    private:
        mutable std::string name;

    protected:
        OrderingEntity* entity;
};

inline OrderingHeuristic::OrderingHeuristic()
    :
        entity(NULL)
{
}

inline OrderingHeuristic::~OrderingHeuristic()
{
}

inline void OrderingHeuristic::setEntity(OrderingEntity* entity)
{
    this->entity = entity;
}

inline bool OrderingHeuristic::isFixedOrder() const
{
    return true;
}

inline const std::string& OrderingHeuristic::getName() const
{
    if(name == "")
        name = toString();
    return name;
}

inline const std::string& OrderingHeuristic::getEntityName() const
{
    return getName();
}

#endif
