// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "OrderingHeuristicStore.hpp"
#include "FirstHeuristic.hpp"
#include "SizeHeuristic.hpp"
#include "EstimateHeuristic.hpp"
#include "RadialHeuristic.hpp"
#include "AngularHeuristic.hpp"
#include "LinearHeuristic.hpp"
#include "PolyPerf.hpp"
#include "fastbits.hpp"
#include "Constants.hpp"
#include "util.hpp"
#include "boost/lexical_cast.hpp"
#include "boost/tokenizer.hpp"
#include "makeArray.hpp"
#include <sstream>
#include <stdexcept>

using namespace std;

void OrderingHeuristicStore::setNumMobilePieces(int numMobilePieces)
{
    if(this->numMobilePieces != 0)
    {
        ostringstream errMsg;
        errMsg << "***OrderingOrderingHeuristicStore::setNumMobilePieces(): this method can only be called once.";
        throw runtime_error(errMsg.str());
    }
    this->numMobilePieces = numMobilePieces;
    this->heuristicList = initArray<OrderingHeuristic*>(numMobilePieces+1, NULL);
}

OrderingHeuristicStore::~OrderingHeuristicStore()
{
    if(heuristicList)
    {
        for(int i = 1; i <= numMobilePieces; ++i)
            delete heuristicList[i];
        deleteArray(heuristicList, 0);
    }

    for(EntityMap::iterator i = entityMap.begin(); i != entityMap.end(); ++i)
        delete i->second;
}

void OrderingHeuristicStore::setHeuristic(int nrp, const std::string& def)
{
    using boost::lexical_cast;

    typedef boost::tokenizer<boost::char_separator<char> > Tokenizer;
    typedef boost::char_separator<char> CharSeparator;

    string name = def;
    string args = "";

    int atSymbol = def.find('@');
    if(atSymbol != -1)
    {
        name = def.substr(0, atSymbol);
        args = def.substr(atSymbol + 1);
    }


    if(args != "" && (name == "s" || name == "e" || name == "f" || name == "x" || name == "X" || name == "y" || name == "Y" || name == "z" || name == "Z" || name == "xyz"))
    {
            ostringstream errMsg;
            errMsg <<
                "***Ordering Heuristic Configuration Syntax Error:  '" << def << "'\n\n"
                "Ordering heuristic " << name << " takes no arguments.\n";

            throw runtime_error(errMsg.str());
    }

    if(name == "s")
    {
        storeHeuristic(nrp, new SizeHeuristic());
        return;
    }

    if(name == "e")
    {
        storeHeuristic(nrp, new EstimateHeuristic());
        return;
    }

    if(name == "f")
    {
        storeHeuristic(nrp, new FirstHeuristic());
        return;
    }

    if(name == "x")
    {
       storeHeuristic(nrp, new LinearHeuristic(1, 0, 0));
       return;
    }

    if(name == "X")
    {
        storeHeuristic(nrp, new LinearHeuristic(-1, 0, 0));
        return;
    }

    if(name == "y")
    {
        storeHeuristic(nrp, new LinearHeuristic(0, 1, 0));
        return;
    }

    if(name == "Y")
    {
        storeHeuristic(nrp, new LinearHeuristic(0, -1, 0));
        return;
    }

    if(name == "z")
    {
        storeHeuristic(nrp, new LinearHeuristic(0, 0, 1));
        return;
    }

    if(name == "Z")
    {
        storeHeuristic(nrp, new LinearHeuristic(0, 0, -1));
        return;
    }

    if(name == "xyz")
    {
        storeHeuristic(nrp, new LinearHeuristic(1, 1, 1));
        return;
    }

    Tokenizer argTok(args, CharSeparator(","));
    int c = countAll(argTok.begin(), argTok.end());
    Tokenizer::iterator i = argTok.begin();

    if(name == "l")
    {
        if(c != 3)
        {
            ostringstream errMsg;
            errMsg <<
                "***Linear Ordering Heuristic Configuration Syntax Error:  '" << def << "'\n\n"
                "The args for a linear ordering heuristic configuration settings\n"
                "must be three comma separated floating point numbers.\n";

            throw runtime_error(errMsg.str());
        }

        double a = lexical_cast<double>(*i++);
        double b = lexical_cast<double>(*i++);
        double c = lexical_cast<double>(*i++);
        storeHeuristic(nrp, new LinearHeuristic(a, b, c));
        return;
    }

    if(name == "a" || name == "A")
    {
        double initialAngle = 0.0;
        if(i != argTok.end())
            initialAngle = lexical_cast<double>(*i++) / 360.0 * TAU;

        double xc = defaultXc;
        if(i != argTok.end())
            xc = lexical_cast<double>(*i++);

        double yc = defaultYc;
        if(i != argTok.end())
            yc = lexical_cast<double>(*i++);

        storeHeuristic(nrp, new AngularHeuristic(initialAngle, xc, yc, name == "A"));
        return;
    }

    if(name == "R")
    {
        if(c > 3)
        {
            ostringstream errMsg;
            errMsg <<
                "***Radial Ordering Heuristic Configuration Syntax Error:  '" << def << "'\n\n"
                "The args for radial ordering heuristic configuration settings\n"
                "must be empty or three comma separated floating point numbers\n"
                "specifying the coordinates from which radial distance is to be\n"
                "measured.\n";

            throw runtime_error(errMsg.str());
        }

        double xc = defaultXc;
        if(i != argTok.end())
            xc = lexical_cast<double>(*i++);

        double yc = defaultYc;
        if(i != argTok.end())
            yc = lexical_cast<double>(*i++);

        double zc = defaultZc;
        if(i != argTok.end())
            zc = lexical_cast<double>(*i++);

        storeHeuristic(nrp, new RadialHeuristic(xc, yc, zc));
        return;
    }

    ostringstream errMsg;
    errMsg <<
        "***Unknown Ordering Heuristic:  '" << def << "'\n\n"
        "Known heuristics are:  a, A, e, f, R, s, x, X, y, Y, z, Z, xyz, l.\n";

    throw runtime_error(errMsg.str());
}


void OrderingHeuristicStore::storeHeuristic(int nrp, OrderingHeuristic* oh)
{
    //std::cout << "heuristicList[" << nrp << "]=" << (*oh) << std::endl;
    heuristicList[nrp] = oh;
    OrderingEntity* entity = genEntity(oh);
    oh->setEntity(entity);
    if(nrp <= fila)
    {
        if(nof && oh->getName() != "f")
            need6Bit = true;
        filaEntitySet.insert(entity);
    }
}


OrderingEntity* OrderingHeuristicStore::genEntity(OrderingHeuristic* oh)
{
    const std::string& entityName = oh->getEntityName();
    EntityMap::iterator i = entityMap.find(entityName);

    if(i != entityMap.end())
        return i->second;

    OrderingEntity* oe = new OrderingEntity(
            oh,
            dlxRoot,
            numMobileShapes,
            remainingShapeList,
            filaGrid,
            nof,
            entityName != "f");

    entityMap[entityName] = oe;
    return oe;
}

void OrderingHeuristicStore::loadImages(int nrp, int numFilaGridPoints)
{
    this->numFilaGridPoints = numFilaGridPoints;

    // If nof is requested, initialize neighborOccupancy
    //
    gridbitfield_t neighborOccupancy[maxFilaGridPoints][64];
    if(nof)
        initNeighborOccupancy(neighborOccupancy);

    for(EntitySet::iterator i = filaEntitySet.begin(); i != filaEntitySet.end(); ++i)
        (*i)->loadImages(nrp, numFilaGridPoints, neighborOccupancy);
}

void OrderingHeuristicStore::initNeighborOccupancy(gridbitfield_t neighborOccupancy[][64])
{
    /** The getIlsi methods basically convert the occupancy states of the
     ** neighbors of some GridPoint g down to a 6 bit number (for indexing).
     ** The neighborOccupancy array basically does the reverse:  translating
     ** an ILSI back to a bit mask corresponding to the set bits of the ILSI.
     ** This information is used to translate ilsi numbers into a conflict
     ** mask that can be used to filter images during the image list setup.
     ** Which is why neighborOccupancy is stack-allocated by the caller --
     ** it's only for setup.
     **
     **     neighborOccupancy[63] = gp->nbr_x | gp->nbr_y | gp->nbr_z | gp->nbr_X | gp->nbr_Y | gp->nbr_Z;
     **     neighborOccupancy[62] =     0     | gp->nbr_y | gp->nbr_z | gp->nbr_X | gp->nbr_Y | gp->nbr_Z;
     **     neighborOccupancy[61] = gp->nbr_x |     0     | gp->nbr_z | gp->nbr_X | gp->nbr_Y | gp->nbr_Z;
     **     neighborOccupancy[60] =     0     |     0     | gp->nbr_z | gp->nbr_X | gp->nbr_Y | gp->nbr_Z;
     **     neighborOccupancy[59] = gp->nbr_x | gp->nbr_y |     0     | gp->nbr_X | gp->nbr_Y | gp->nbr_Z;
     **     .
     **     .
     **     .
     **     neighborOccupancy[0]  = 0;
     **     etc.
     **
     ** Any image with a layoutMask that overlaps with neighborOccupancy[i] is not
     ** allowed in the i-th image list.
     **/

    for(int g = 0; g < numFilaGridPoints; ++g)
    {
        const GridPoint* gp = filaGrid[g];
        if(need6Bit)
        {
            for(int ilsi = 0; ilsi < 64; ++ilsi)
            {
                gridbitfield_t& no = neighborOccupancy[gp->bitNum][ilsi];
                no = 0;
                if(ilsi&1)
                    no |= gp->nbr_z;
                if(ilsi&2)
                    no |= gp->nbr_y;
                if(ilsi&4)
                    no |= gp->nbr_x;
                if(ilsi&8)
                    no |= gp->nbr_Z;
                if(ilsi&16)
                    no |= gp->nbr_Y;
                if(ilsi&32)
                    no |= gp->nbr_X;
            }
        }
        else  // TODO:  not sure how much time this really saves.  Is it worth the special case?
        {
            for(int ilsi = 56; ilsi < 64; ++ilsi)
            {
                gridbitfield_t& no = neighborOccupancy[gp->bitNum][ilsi];
                no = 0;
                if(ilsi&1)
                    no |= gp->nbr_z;
                if(ilsi&2)
                    no |= gp->nbr_y;
                if(ilsi&4)
                    no |= gp->nbr_x;
                //std::cout << "neighborOccupancy[" << gp->bitNum << "][" << ilsi << "]=" << no << std::endl;
            }
        }
    }
}


void OrderingHeuristicStore::clearImages()
{
    for(EntitySet::iterator i = filaEntitySet.begin(); i != filaEntitySet.end(); ++i)
        (*i)->clearImages();
}
