// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FASTBITS_HPP
#define FASTBITS_HPP

#include <stdint.h>

namespace FastBits
{
    // This file contains some rediculous-looking inline functions supporting
    // fast iteration over bits set in unsigned integers of various precision.
    // Sample code looks like this:
    //
    //     unsigned int x = 0x800100F1;
    //     for(int bit = lowestSetBit(x); bit != -1; bit = nextLowestSetBit(x, bit))
    //         cout << bit << endl;
    //
    // which outputs:
    //
    //      0
    //      4
    //      5
    //      6
    //      7
    //     16
    //     31
    //
    // There's also routines for counting set bits:
    //
    //     unsigned int x = 0x800100F1;
    //     cout << countBits(x) << endl;
    //
    // which outputs:
    //     7
    //
    // My countBit routines in particular I know to be at least 2 times faster
    // (in my testing on my I3-4130T with -O3 optimization) than even Gnu's
    // builtin routines __builtin_popcount().  I say "at least" because I only
    // measured total test program run (which was more than twice as fast), but
    // I didn't try to measure how much time was due to the common for-loop
    // construct.  I thought the popcount routines were supposed to be mapped
    // to a single silicon-based instruction, so I'm not sure how my routines
    // are faster, but they certainly appear to be.

    extern const int         lowestSetBitMap[256];

    extern const int         highestSetBitMap[256];

    extern const int         bitCount[256];

    extern const uint8_t     upperMask8  [  9];
    extern const uint16_t    upperMask16 [ 17];
    extern const uint32_t    upperMask32 [ 33];
    extern const uint64_t    upperMask64 [ 65];

    extern const uint8_t     lowerMask8  [  9];
    extern const uint16_t    lowerMask16 [ 17];
    extern const uint32_t    lowerMask32 [ 33];
    extern const uint64_t    lowerMask64 [ 65];


    template<typename T> T upperMask(int bit) = delete;  // Purposefully left unimplemented.

    template<>    inline uint8_t     upperMask<uint8_t>    (int bit) { return upperMask8  [bit]; }
    template<>    inline uint16_t    upperMask<uint16_t>   (int bit) { return upperMask16 [bit]; }
    template<>    inline uint32_t    upperMask<uint32_t>   (int bit) { return upperMask32 [bit]; }
    template<>    inline uint64_t    upperMask<uint64_t>   (int bit) { return upperMask64 [bit]; }

    const uint16_t m16_01 = 0x00ffu;
    const uint16_t m16_10 = 0xff00u;


    const uint32_t m32_0001 = 0x000000ffu;
    const uint32_t m32_0010 = 0x0000ff00u;
    const uint32_t m32_0100 = 0x00ff0000u;
    const uint32_t m32_1000 = 0xff000000u;

    const uint32_t m32_0011 = 0x0000ffffu;
    const uint32_t m32_1100 = 0xffff0000u;


    const uint64_t m64_00000001 = 0x00000000000000ffull;
    const uint64_t m64_00000010 = 0x000000000000ff00ull;
    const uint64_t m64_00000100 = 0x0000000000ff0000ull;
    const uint64_t m64_00001000 = 0x00000000ff000000ull;
    const uint64_t m64_00010000 = 0x000000ff00000000ull;
    const uint64_t m64_00100000 = 0x0000ff0000000000ull;
    const uint64_t m64_01000000 = 0x00ff000000000000ull;
    const uint64_t m64_10000000 = 0xff00000000000000ull;

    const uint64_t m64_00000011 = 0x000000000000ffffull;
    const uint64_t m64_00001100 = 0x00000000ffff0000ull;
    const uint64_t m64_00110000 = 0x0000ffff00000000ull;
    const uint64_t m64_11000000 = 0xffff000000000000ull;

    const uint64_t m64_00001111 = 0x00000000ffffffffull;
    const uint64_t m64_11110000 = 0xffffffff00000000ull;

    inline int countBits(uint8_t bitField)
    {
        return bitCount[bitField];
    }

    inline int countBits(uint16_t bitField)
    {
        return
            bitCount[(bitField & m16_01)     ] +
            bitCount[(bitField & m16_10) >> 8];
    }

    inline int countBits(uint32_t bitField)
    {
#ifdef USE_GNU_BUILTINS
        return __builtin_popcount(bitField);
#else
        return
            bitCount[(bitField & m32_0001)      ] +
            bitCount[(bitField & m32_0010) >>  8] +
            bitCount[(bitField & m32_0100) >> 16] +
            bitCount[(bitField & m32_1000) >> 24];
#endif
    }

    inline int countBits(uint64_t bitField)
    {
#ifdef USE_GNU_BUILTINS
        return __builtin_popcountll(bitField);
#else
        return
            bitCount[(bitField & m64_00000001)      ] +
            bitCount[(bitField & m64_00000010) >>  8] +
            bitCount[(bitField & m64_00000100) >> 16] +
            bitCount[(bitField & m64_00001000) >> 24] +
            bitCount[(bitField & m64_00010000) >> 32] +
            bitCount[(bitField & m64_00100000) >> 40] +
            bitCount[(bitField & m64_01000000) >> 48] +
            bitCount[(bitField & m64_10000000) >> 56];
#endif
    }

    inline int lowestSetBit(uint8_t  bitField)
    {
        return lowestSetBitMap[bitField];
    }

    inline int lowestSetBit(uint16_t bitField)
    {
        if(bitField & m16_01)
            return lowestSetBitMap[ bitField & m16_01      ];
        else
            return lowestSetBitMap[(bitField & m16_10) >> 8] + 8;
    }

    inline int lowestSetBit(uint32_t bitField)
    {
#ifdef USE_GNU_BUILTINS
        if(bitField)
            return __builtin_ctz(bitField);
        else
            return -1;
#else
        if(bitField & m32_0011)
            if(bitField & m32_0001)
                return lowestSetBitMap[ bitField & m32_0001       ];
            else
                return lowestSetBitMap[(bitField & m32_0010) >>  8] + 8;
        else
            if(bitField & m32_0100)
                return lowestSetBitMap[(bitField & m32_0100) >> 16] + 16;
            else
                return lowestSetBitMap[ bitField             >> 24] + 24;
#endif
    }

    inline int lowestSetBit(uint64_t bitField)
    {
#ifdef USE_GNU_BUILTINS
        if(bitField)
            return __builtin_ctzll(bitField);
        else
            return -1;
#else
        if(bitField & m64_00001111)
            if(bitField & m64_00000011)
                if(bitField & m64_00000001)
                    return lowestSetBitMap[ bitField & m64_00000001       ];
                else
                    return lowestSetBitMap[(bitField & m64_00000010) >>  8] + 8;
            else
                if(bitField & m64_00000100)
                    return lowestSetBitMap[(bitField & m64_00000100) >> 16] + 16;
                else
                    return lowestSetBitMap[(bitField & m64_00001000) >> 24] + 24;
        else
            if(bitField & m64_00110000)
                if(bitField & m64_00010000)
                    return lowestSetBitMap[(bitField & m64_00010000) >> 32] + 32;
                else
                    return lowestSetBitMap[(bitField & m64_00100000) >> 40] + 40;
            else
                if(bitField & m64_01000000)
                    return lowestSetBitMap[(bitField & m64_01000000) >> 48] + 48;
                else
                    return lowestSetBitMap[ bitField                 >> 56] + 56;
#endif
    }

    inline int nextLowestSetBit(uint8_t bitField, int limit)
    {
        return lowestSetBitMap[bitField & upperMask8[limit+1]];
    }


    inline int nextLowestSetBit(uint16_t bitField, int limit)
    {
        bitField = bitField & upperMask16[limit+1];

        switch((limit+1)>>3)
        {
            case 0:
                if(bitField & m16_01)
                    return lowestSetBitMap[bitField & m16_01];
            case 1:
                if(bitField & m16_10)
                    return lowestSetBitMap[(bitField & m16_10) >>   8] +   8;
            default:
                return -1;
        }
    }


    inline int nextLowestSetBit(uint32_t bitField, int limit)
    {
        bitField = bitField & upperMask32[limit+1];
#ifdef USE_GNU_BUILTINS
        if(bitField)
            return __builtin_ctz(bitField);
        else
            return -1;
#else
        switch((limit+1)>>3)
        {
            case 0:
                if(bitField & m32_0001)
                    return lowestSetBitMap[bitField & m32_0001];
            case 1:
                if(bitField & m32_0010)
                    return lowestSetBitMap[(bitField & m32_0010) >>   8] +   8;
            case 2:
                if(bitField & m32_0100)
                    return lowestSetBitMap[(bitField & m32_0100) >>  16] +  16;
            case 3:
                if(bitField & m32_1000)
                    return lowestSetBitMap[(bitField & m32_1000) >>  24] +  24;
            default:
                return -1;
        }
#endif
    }


    inline int nextLowestSetBit(uint64_t bitField, int limit)
    {
        bitField = bitField & upperMask64[limit+1];
#ifdef USE_GNU_BUILTINS
        if(bitField)
            return __builtin_ctzll(bitField);
        else
            return -1;
#else
        switch((limit+1)>>3)
        {
            case 0:
                if(bitField & m64_00000001)
                    return lowestSetBitMap[bitField & m64_00000001];
            case 1:
                if(bitField & m64_00000010)
                    return lowestSetBitMap[(bitField & m64_00000010) >>   8] +   8;
            case 2:
                if(bitField & m64_00000100)
                    return lowestSetBitMap[(bitField & m64_00000100) >>  16] +  16;
            case 3:
                if(bitField & m64_00001000)
                    return lowestSetBitMap[(bitField & m64_00001000) >>  24] +  24;
            case 4:
                if(bitField & m64_00010000)
                    return lowestSetBitMap[(bitField & m64_00010000) >>  32] +  32;
            case 5:
                if(bitField & m64_00100000)
                    return lowestSetBitMap[(bitField & m64_00100000) >>  40] +  40;
            case 6:
                if(bitField & m64_01000000)
                    return lowestSetBitMap[(bitField & m64_01000000) >>  48] +  48;
            case 7:
                if(bitField & m64_10000000)
                    return lowestSetBitMap[(bitField & m64_10000000) >>  56] +  56;
            default:
                return -1;
        }
#endif
    }


    inline int highestSetBit(uint8_t  bitField)
    {
        return highestSetBitMap[bitField];
    }

    inline int highestSetBit(uint16_t bitField)
    {
        if(bitField & m16_10)
            return highestSetBitMap[bitField >> 8] + 8;
        else
            return highestSetBitMap[bitField     ];
    }

    inline int highestSetBit(uint32_t bitField)
    {
        if(bitField & m32_1100)
            if(bitField & m32_1000)
                return highestSetBitMap[bitField >> 24] + 24;
            else
                return highestSetBitMap[bitField >> 16] + 16;
        else
            if(bitField & m32_0010)
                return highestSetBitMap[bitField >>  8] +  8;
            else
                return highestSetBitMap[bitField      ];
    }

    inline int highestSetBit(uint64_t bitField)
    {
        if(bitField & m64_11110000)
            if(bitField & m64_11000000)
                if(bitField & m64_10000000)
                    return highestSetBitMap[bitField >> 56] + 56;
                else
                    return highestSetBitMap[bitField >> 48] + 48;
            else
                if(bitField & m64_00100000)
                    return highestSetBitMap[bitField >> 40] + 40;
                else
                    return highestSetBitMap[bitField >> 32] + 32;
        else
            if(bitField & m64_00001100)
                if(bitField & m64_00001000)
                    return highestSetBitMap[bitField >> 24] + 24;
                else
                    return highestSetBitMap[bitField >> 16] + 16;
            else
                if(bitField & m64_00000010)
                    return highestSetBitMap[bitField >>  8] +  8;
                else
                    return highestSetBitMap[bitField      ];
    }

    inline int nextHighestSetBit(uint8_t bitField, int limit)
    {
        return highestSetBitMap[bitField & lowerMask8[limit]];
    }


    inline int nextHighestSetBit(uint16_t bitField, int limit)
    {
        bitField = bitField & lowerMask16[limit];

        switch((limit-1)>>3)
        {
            case 1:
                if(bitField & m16_10)
                    return highestSetBitMap[bitField >> 8] + 8;
            case 0:
                if(bitField & m16_01)
                    return highestSetBitMap[bitField     ];
            default:
                return -1;
        }
    }


    inline int nextHighestSetBit(uint32_t bitField, int limit)
    {
        bitField = bitField & lowerMask32[limit];

        switch((limit-1)>>3)
        {
            case 3:
                if(bitField & m32_1000)
                    return highestSetBitMap[bitField >> 24] + 24;
            case 2:
                if(bitField & m32_0100)
                    return highestSetBitMap[bitField >> 16] + 16;
            case 1:
                if(bitField & m32_0010)
                    return highestSetBitMap[bitField >>  8] +  8;
            case 0:
                if(bitField & m32_0001)
                    return highestSetBitMap[bitField      ];
            default:
                return -1;
        }
    }


    inline int nextHighestSetBit(uint64_t bitField, int limit)
    {
        bitField = bitField & lowerMask64[limit];

        switch((limit-1)>>3)
        {
            case 7:
                if(bitField & m64_10000000)
                    return highestSetBitMap[bitField >> 56] + 56;
            case 6:
                if(bitField & m64_01000000)
                    return highestSetBitMap[bitField >> 48] + 48;
            case 5:
                if(bitField & m64_00100000)
                    return highestSetBitMap[bitField >> 40] + 40;
            case 4:
                if(bitField & m64_00010000)
                    return highestSetBitMap[bitField >> 32] + 32;
            case 3:
                if(bitField & m64_00001000)
                    return highestSetBitMap[bitField >> 24] + 24;
            case 2:
                if(bitField & m64_00000100)
                    return highestSetBitMap[bitField >> 16] + 16;
            case 1:
                if(bitField & m64_00000010)
                    return highestSetBitMap[bitField >>  8] +  8;
            case 0:
                if(bitField & m64_00000001)
                    return highestSetBitMap[bitField      ];
            default:
                return -1;
        }
    }



    // Only some versions of GNU compilers can compile this __uint128_t guff.
    // Leaving out all reference to it unless GRIDBITFIELD_SIZE preprocessor
    // definition is set to 128
    //
    #if GRIDBITFIELD_SIZE == 128
        extern const __uint128_t upperMask128[129];
        extern const __uint128_t lowerMask128[129];
        template<>    inline __uint128_t upperMask<__uint128_t>(int bit) { return upperMask128[bit]; }

        // Currently there's no way to write a string literal for 128 bit
        // integers, so I instead build them up from previously defined constants
        // using casts, shifts, and adds.

        const __uint128_t m128_0000000000000001 = m64_00000001;
        const __uint128_t m128_0000000000000010 = m64_00000010;
        const __uint128_t m128_0000000000000100 = m64_00000100;
        const __uint128_t m128_0000000000001000 = m64_00001000;
        const __uint128_t m128_0000000000010000 = m64_00010000;
        const __uint128_t m128_0000000000100000 = m64_00100000;
        const __uint128_t m128_0000000001000000 = m64_01000000;
        const __uint128_t m128_0000000010000000 = m64_10000000;
        const __uint128_t m128_0000000100000000 = ((__uint128_t) m64_00000001 << 64);
        const __uint128_t m128_0000001000000000 = ((__uint128_t) m64_00000010 << 64);
        const __uint128_t m128_0000010000000000 = ((__uint128_t) m64_00000100 << 64);
        const __uint128_t m128_0000100000000000 = ((__uint128_t) m64_00001000 << 64);
        const __uint128_t m128_0001000000000000 = ((__uint128_t) m64_00010000 << 64);
        const __uint128_t m128_0010000000000000 = ((__uint128_t) m64_00100000 << 64);
        const __uint128_t m128_0100000000000000 = ((__uint128_t) m64_01000000 << 64);
        const __uint128_t m128_1000000000000000 = ((__uint128_t) m64_10000000 << 64);

        const __uint128_t m128_0000000000000011 = m128_0000000000000001 + m128_0000000000000010;
        const __uint128_t m128_0000000000001100 = m128_0000000000000100 + m128_0000000000001000;
        const __uint128_t m128_0000000000110000 = m128_0000000000010000 + m128_0000000000100000;
        const __uint128_t m128_0000000011000000 = m128_0000000001000000 + m128_0000000010000000;
        const __uint128_t m128_0000001100000000 = m128_0000000100000000 + m128_0000001000000000;
        const __uint128_t m128_0000110000000000 = m128_0000010000000000 + m128_0000100000000000;
        const __uint128_t m128_0011000000000000 = m128_0001000000000000 + m128_0010000000000000;
        const __uint128_t m128_1100000000000000 = m128_0100000000000000 + m128_1000000000000000;

        const __uint128_t m128_0000000000001111 = m128_0000000000000011 + m128_0000000000001100;
        const __uint128_t m128_0000000011110000 = m128_0000000000110000 + m128_0000000011000000;
        const __uint128_t m128_0000111100000000 = m128_0000001100000000 + m128_0000110000000000;
        const __uint128_t m128_1111000000000000 = m128_0011000000000000 + m128_1100000000000000;

        const __uint128_t m128_0000000011111111 = m128_0000000000001111 + m128_0000000011110000;
        const __uint128_t m128_1111111100000000 = m128_0000111100000000 + m128_1111000000000000;

        inline int countBits(__uint128_t bitField)
        {
            #ifdef USE_GNU_BUILTINS
                return __builtin_popcountll(bitField)
                    + __builtin_popcountll(bitField >> 64);
            #else
                return
                    bitCount[(bitField & m128_0000000000000001)       ] +
                    bitCount[(bitField & m128_0000000000000010) >>   8] +
                    bitCount[(bitField & m128_0000000000000100) >>  16] +
                    bitCount[(bitField & m128_0000000000001000) >>  24] +
                    bitCount[(bitField & m128_0000000000010000) >>  32] +
                    bitCount[(bitField & m128_0000000000100000) >>  40] +
                    bitCount[(bitField & m128_0000000001000000) >>  48] +
                    bitCount[(bitField & m128_0000000010000000) >>  56] +
                    bitCount[(bitField & m128_0000000100000000) >>  64] +
                    bitCount[(bitField & m128_0000001000000000) >>  72] +
                    bitCount[(bitField & m128_0000010000000000) >>  80] +
                    bitCount[(bitField & m128_0000100000000000) >>  88] +
                    bitCount[(bitField & m128_0001000000000000) >>  96] +
                    bitCount[(bitField & m128_0010000000000000) >> 104] +
                    bitCount[(bitField & m128_0100000000000000) >> 112] +
                    bitCount[(bitField & m128_1000000000000000) >> 120];
            #endif
        }

        inline int lowestSetBit(__uint128_t bitField)
        {
            #ifdef USE_GNU_BUILTINS
                if(bitField & m128_0000000011111111)
                    return __builtin_ctzll(bitField);
                else if(bitField)
                    return __builtin_ctzll(bitField >> 64) + 64;
                else
                    return -1;
            #else
                if(bitField & m128_0000000011111111)
                    if(bitField & m128_0000000000001111)
                        if(bitField & m128_0000000000000011)
                            if(bitField & m128_0000000000000001)
                                return lowestSetBitMap[ bitField & m128_0000000000000001        ];
                            else
                                return lowestSetBitMap[(bitField & m128_0000000000000010) >>   8] + 8;
                        else
                            if(bitField & m128_0000000000000100)
                                return lowestSetBitMap[(bitField & m128_0000000000000100) >>  16] + 16;
                            else
                                return lowestSetBitMap[(bitField & m128_0000000000001000) >>  24] + 24;
                    else
                        if(bitField & m128_0000000000110000)
                            if(bitField & m128_0000000000010000)
                                return lowestSetBitMap[(bitField & m128_0000000000010000) >>  32] + 32;
                            else
                                return lowestSetBitMap[(bitField & m128_0000000000100000) >>  40] + 40;
                        else
                            if(bitField & m128_0000000001000000)
                                return lowestSetBitMap[(bitField & m128_0000000001000000) >>  48] + 48;
                            else
                                return lowestSetBitMap[(bitField & m128_0000000010000000) >>  56] + 56;
                else
                    if(bitField & m128_0000111100000000)
                        if(bitField & m128_0000001100000000)
                            if(bitField & m128_0000000100000000)
                                return lowestSetBitMap[(bitField & m128_0000000100000000) >>  64] + 64;
                            else
                                return lowestSetBitMap[(bitField & m128_0000001000000000) >>  72] + 72;
                        else
                            if(bitField & m128_0000010000000000)
                                return lowestSetBitMap[(bitField & m128_0000010000000000) >>  80] + 80;
                            else
                                return lowestSetBitMap[(bitField & m128_0000100000000000) >>  88] + 88;
                    else
                        if(bitField & m128_0011000000000000)
                            if(bitField & m128_0001000000000000)
                                return lowestSetBitMap[(bitField & m128_0001000000000000) >>  96] + 96;
                            else
                                return lowestSetBitMap[(bitField & m128_0010000000000000) >> 104] + 104;
                        else
                            if(bitField & m128_0100000000000000)
                                return lowestSetBitMap[(bitField & m128_0100000000000000) >> 112] + 112;
                            else
                                return lowestSetBitMap[(bitField & m128_1000000000000000) >> 120] + 120;
            #endif
        }

        inline int nextLowestSetBit(__uint128_t bitField, int limit)
        {
            bitField = bitField & upperMask128[limit+1];
            #ifdef USE_GNU_BUILTINS
                if(bitField & m128_0000000011111111)
                    return __builtin_ctzll(bitField);
                else if(bitField)
                    return __builtin_ctzll(bitField >> 64) + 64;
                else
                    return -1;
            #else
                switch((limit+1)>>3)
                {
                    case 0:
                        if(bitField & m128_0000000000000001)
                            return lowestSetBitMap[bitField & m128_0000000000000001];
                    case 1:
                        if(bitField & m128_0000000000000010)
                            return lowestSetBitMap[(bitField & m128_0000000000000010) >>   8] +   8;
                    case 2:
                        if(bitField & m128_0000000000000100)
                            return lowestSetBitMap[(bitField & m128_0000000000000100) >>  16] +  16;
                    case 3:
                        if(bitField & m128_0000000000001000)
                            return lowestSetBitMap[(bitField & m128_0000000000001000) >>  24] +  24;
                    case 4:
                        if(bitField & m128_0000000000010000)
                            return lowestSetBitMap[(bitField & m128_0000000000010000) >>  32] +  32;
                    case 5:
                        if(bitField & m128_0000000000100000)
                            return lowestSetBitMap[(bitField & m128_0000000000100000) >>  40] +  40;
                    case 6:
                        if(bitField & m128_0000000001000000)
                            return lowestSetBitMap[(bitField & m128_0000000001000000) >>  48] +  48;
                    case 7:
                        if(bitField & m128_0000000010000000)
                            return lowestSetBitMap[(bitField & m128_0000000010000000) >>  56] +  56;
                    case 8:
                        if(bitField & m128_0000000100000000)
                            return lowestSetBitMap[(bitField & m128_0000000100000000) >>  64] +  64;
                    case 9:
                        if(bitField & m128_0000001000000000)
                            return lowestSetBitMap[(bitField & m128_0000001000000000) >>  72] +  72;
                    case 10:
                        if(bitField & m128_0000010000000000)
                            return lowestSetBitMap[(bitField & m128_0000010000000000) >>  80] +  80;
                    case 11:
                        if(bitField & m128_0000100000000000)
                            return lowestSetBitMap[(bitField & m128_0000100000000000) >>  88] +  88;
                    case 12:
                        if(bitField & m128_0001000000000000)
                            return lowestSetBitMap[(bitField & m128_0001000000000000) >>  96] +  96;
                    case 13:
                        if(bitField & m128_0010000000000000)
                            return lowestSetBitMap[(bitField & m128_0010000000000000) >> 104] + 104;
                    case 14:
                        if(bitField & m128_0100000000000000)
                            return lowestSetBitMap[(bitField & m128_0100000000000000) >> 112] + 112;
                    case 15:
                        if(bitField & m128_1000000000000000)
                            return lowestSetBitMap[(bitField & m128_1000000000000000) >> 120] + 120;
                    default:
                        return -1;
                }
            #endif
        }

        inline int highestSetBit(__uint128_t bitField)
        {
            if(bitField & m128_1111111100000000)
                if(bitField & m128_1111000000000000)
                    if(bitField & m128_1100000000000000)
                        if(bitField & m128_1000000000000000)
                            return highestSetBitMap[bitField >> 120] + 120;
                        else
                            return highestSetBitMap[bitField >> 112] + 112;
                    else
                        if(bitField & m128_0010000000000000)
                            return highestSetBitMap[bitField >> 104] + 104;
                        else
                            return highestSetBitMap[bitField >>  96] +  96;
                else
                    if(bitField & m128_0000110000000000)
                        if(bitField & m128_0000100000000000)
                            return highestSetBitMap[bitField >>  88] +  88;
                        else
                            return highestSetBitMap[bitField >>  80] +  80;
                    else
                        if(bitField & m128_0000001000000000)
                            return highestSetBitMap[bitField >>  72] +  72;
                        else
                            return highestSetBitMap[bitField >>  64] +  64;
            else
                if(bitField & m128_0000000011110000)
                    if(bitField & m128_0000000011000000)
                        if(bitField & m128_0000000010000000)
                            return highestSetBitMap[bitField >>  56] +  56;
                        else
                            return highestSetBitMap[bitField >>  48] +  48;
                    else
                        if(bitField & m128_0000000000100000)
                            return highestSetBitMap[bitField >>  40] +  40;
                        else
                            return highestSetBitMap[bitField >>  32] +  32;
                else
                    if(bitField & m128_0000000000001100)
                        if(bitField & m128_0000000000001000)
                            return highestSetBitMap[bitField >>  24] +  24;
                        else
                            return highestSetBitMap[bitField >>  16] +  16;
                    else
                        if(bitField & m128_0000000000000010)
                            return highestSetBitMap[bitField >>   8] +   8;
                        else
                            return highestSetBitMap[bitField       ];
        }


        inline int nextHighestSetBit(__uint128_t bitField, int limit)
        {
            bitField = bitField & lowerMask128[limit];

            switch((limit-1)>>3)
            {
                case 15:
                    if(bitField & m128_1000000000000000)
                        return highestSetBitMap[bitField >> 120] + 120;
                case 14:
                    if(bitField & m128_0100000000000000)
                        return highestSetBitMap[bitField >> 112] + 112;
                case 13:
                    if(bitField & m128_0010000000000000)
                        return highestSetBitMap[bitField >> 104] + 104;
                case 12:
                    if(bitField & m128_0001000000000000)
                        return highestSetBitMap[bitField >>  96] +  96;
                case 11:
                    if(bitField & m128_0000100000000000)
                        return highestSetBitMap[bitField >>  88] +  88;
                case 10:
                    if(bitField & m128_0000010000000000)
                        return highestSetBitMap[bitField >>  80] +  80;
                case 9:
                    if(bitField & m128_0000001000000000)
                        return highestSetBitMap[bitField >>  72] +  72;
                case 8:
                    if(bitField & m128_0000000100000000)
                        return highestSetBitMap[bitField >>  64] +  64;
                case 7:
                    if(bitField & m128_0000000010000000)
                        return highestSetBitMap[bitField >>  56] +  56;
                case 6:
                    if(bitField & m128_0000000001000000)
                        return highestSetBitMap[bitField >>  48] +  48;
                case 5:
                    if(bitField & m128_0000000000100000)
                        return highestSetBitMap[bitField >>  40] +  40;
                case 4:
                    if(bitField & m128_0000000000010000)
                        return highestSetBitMap[bitField >>  32] +  32;
                case 3:
                    if(bitField & m128_0000000000001000)
                        return highestSetBitMap[bitField >>  24] +  24;
                case 2:
                    if(bitField & m128_0000000000000100)
                        return highestSetBitMap[bitField >>  16] +  16;
                case 1:
                    if(bitField & m128_0000000000000010)
                        return highestSetBitMap[bitField >>   8] +   8;
                case 0:
                    if(bitField & m128_0000000000000001)
                        return highestSetBitMap[bitField       ];
                default:
                    return -1;
            }
        }
    #endif
}

#endif
